//
//  viterbiPlugIn.m
//  viterbi
//
//  Created by Dustin O'Connor on 4/30/11.
//  Copyright 2011 umaine. All rights reserved.
//

/* It's highly recommended to use CGL macros instead of changing the current context for plug-ins that perform OpenGL rendering */
#import <OpenGL/CGLMacro.h>

#import "viterbiPlugIn.h"

#include <string>
#include <vector>
#include <map>
#include <iostream>


#define	kQCPlugIn_Name				@"viterbi"
#define	kQCPlugIn_Description		@"viterbi description"


using namespace std;


vector<string> states;
vector<string> observations;
map<string,double> start_probability;
map<string,map<string, double> > transition_probability;
map<string,map<string, double> > emission_probability;

class Tracking {
public:
    double prob;
    vector<string> v_path;
    double v_prob;
    
    Tracking() {
        prob = 0.0;
        v_prob = 0.0;
    }
    
    Tracking(double p, vector<string> & v_pth, double v_p) {
        prob = p;
        v_path = v_pth;
        v_prob = v_p;
    }
};
    //////////////////////////////////////////////////////////////////////////
    //////////////////////////////////////////////////////////////////////////

    void init_variables(string s1, string s2, string s3, string s4, string o1, string o2, string o3, string o4, double sprob1, double sprob2,  double sprob3, double sprob4) {
        
     states.erase (states.begin(),states.end()) ;  
    
    states.push_back(s1);
	states.push_back(s2);
    states.push_back(s3);
    states.push_back(s4);

    observations.erase (observations.begin(),observations.end()) ;  
        
        
    observations.push_back(o1);
    observations.push_back(o2);
    observations.push_back(o3);
     observations.push_back(o4);
  
        start_probability[s1] = sprob1; 
        start_probability[s2] = sprob2;
        start_probability[s3] = sprob3; 
        start_probability[s4] = sprob4;

  
    transition_probability[s1][s1] = .2;
    transition_probability[s1][s2] = .5;
    transition_probability[s1][s3] = .2;
    transition_probability[s1][s4] = .1;
        
    transition_probability[s2][s1] = .1;
   transition_probability[s2][s2] = .2;
   transition_probability[s2][s3] = .5;
   transition_probability[s2][s4] = .2;
        
 transition_probability[s3][s1] = .1;
 transition_probability[s3][s2] = .2;
 transition_probability[s3][s3] = .2;
  transition_probability[s3][s4] = .5;
        
  transition_probability[s4][s1] = .5;
  transition_probability[s4][s2] = .2;
  transition_probability[s4][s3] = .1;
   transition_probability[s4][s4] = .2;

        emission_probability[s1][o1] = .1;
        emission_probability[s1][o2] = .2;
        emission_probability[s1][o3] = .3;
        emission_probability[s1][o4] = .4;
        
        emission_probability[s2][o1] = .2;
        emission_probability[s2][o2] = .1;
        emission_probability[s2][o3] = .2;
        emission_probability[s2][o4] = .5;
        
        emission_probability[s3][o1] = .5;
        emission_probability[s3][o2] = .2;
        emission_probability[s3][o3] = .1;
        emission_probability[s3][o4] = .2;
        
        emission_probability[s4][o1] = .4;
        emission_probability[s4][o2] = .3;
        emission_probability[s4][o3] = .2;
        emission_probability[s4][o4] = .1;

        


  

}

    //////////////////////////////////////////////////////////////////////////
    //////////////////////////////////////////////////////////////////////////

void print_variables(void) {
        // print states
    cout << "States:" << endl;
    for(vector<string>::iterator i=states.begin();i!=states.end();i++) {
        cout << "S: " << (*i) << endl;
    }
        // print observations
    cout << "Observations:" << endl;
    for(vector<string>::iterator i=observations.begin();i!=observations.end();i++) {
        cout << "O: " << (*i) << endl;
    }
    
        // print start probabilities
    cout << "Start probabilities:" << endl;
    for(map<string, double>::iterator i=start_probability.begin();i!=start_probability.end();i++) {
        cout << "S: " << (*i).first << " P: " << (*i).second << endl;
    }
    
        // print transition_probability
    cout << "Transition probabilities:" << endl;
    for(map<string,map<string, double> >::iterator i=transition_probability.begin();i!=transition_probability.end();i++) {
        for(map<string, double>::iterator j=(*i).second.begin();j!=(*i).second.end();j++) {
            cout << "FS: " << (*i).first << " TS: " << (*j).first << " P: " << (*j).second << endl;
        }
    }
    
        // print emission probabilities
    cout << "Emission probabilities:" << endl;
    for(int i=0; i<states.size(); i++) {
        for(int j=0; j<observations.size(); j++) {
            cout << "FS: " << states[i] << " TO: " << observations[j] <<
            " P: " << emission_probability[states[i]][observations[j]] << endl;
        }
    }
}

    //this method compute total probability for observation, most likely viterbi path 
    //and probability of such path
vector<string> forward_viterbi(vector<string> obs, vector<string> states, map<string, double> start_p, 
                     map<string, map<string, double> > trans_p, 
                     map<string, map<string, double> > emit_p) {
    map<string, Tracking> T;
    
    for(vector<string>::iterator state=states.begin(); state!=states.end();state++) {
        vector<string> v_pth;
        v_pth.push_back(*state);
        
        T[*state] = Tracking(start_p[*state], v_pth, start_p[*state]);
    }
    
    for(vector<string>::iterator output=obs.begin(); output!=obs.end();output++) {
        map<string, Tracking> U;
        
        for(vector<string>::iterator next_state=states.begin(); next_state!=states.end(); next_state++) {
            Tracking next_tracker;
            
            for(vector<string>::iterator source_state=states.begin(); source_state!=states.end(); source_state++) {
                Tracking source_tracker = T[*source_state];
                
                double p = emit_p[*source_state][*output]*trans_p[*source_state][*next_state];
                source_tracker.prob *= p;
                source_tracker.v_prob *= p;
                
                next_tracker.prob += source_tracker.prob;
                
                if(source_tracker.v_prob > next_tracker.v_prob) {
                    next_tracker.v_path = source_tracker.v_path;
                    next_tracker.v_path.push_back(*next_state);
                    next_tracker.v_prob = source_tracker.v_prob;
                }
            }
            
            U[*next_state] = next_tracker;
        }
        
        T = U;
    }
    
        // apply sum/max to the final states
    Tracking final_tracker;
    
    for(vector<string>::iterator state=states.begin(); state!=states.end(); state++) {
        Tracking tracker = T[*state];
        
        final_tracker.prob += tracker.prob;
        
        if(tracker.v_prob > final_tracker.v_prob) {
            final_tracker.v_path = tracker.v_path;
            final_tracker.v_prob = tracker.v_prob;
        }
    }
    vector<string> predict;
    predict.erase (predict.begin(),predict.end()); 

        //  cout << "Total probability of the observation sequence: " << final_tracker.prob << endl;
        //    cout << "Probability of the Viterbi path: " << final_tracker.v_prob << endl;
        // cout << "The Viterbi path: " << endl;
    for(vector<string>::iterator state=final_tracker.v_path.begin(); state!=final_tracker.v_path.end(); state++) {
            // cout << "VState: " << *state << endl;
      predict.push_back(*state);
    }
    
    return predict;
}


@implementation viterbiPlugIn

@synthesize path;

@dynamic input_run;

@dynamic input_state_1;
@dynamic input_state_2;
@dynamic input_state_3;
@dynamic input_state_4;



@dynamic input_observation_1;
@dynamic input_observation_2;
@dynamic input_observation_3;
@dynamic input_observation_4;



@dynamic input_state_1_prob;
@dynamic input_state_2_prob;
@dynamic input_state_3_prob;
@dynamic input_state_4_prob;



@dynamic output_v_path;

/*
Here you need to declare the input / output properties as dynamic as Quartz Composer will handle their implementation
@dynamic inputFoo, outputBar;
*/

+ (NSDictionary *)attributes
{
	/*
	Return a dictionary of attributes describing the plug-in (QCPlugInAttributeNameKey, QCPlugInAttributeDescriptionKey...).
	*/
	
	return [NSDictionary dictionaryWithObjectsAndKeys:kQCPlugIn_Name, QCPlugInAttributeNameKey, kQCPlugIn_Description, QCPlugInAttributeDescriptionKey, nil];
}

+ (NSDictionary *)attributesForPropertyPortWithKey:(NSString *)key
{
	/*
	Specify the optional attributes for property based ports (QCPortAttributeNameKey, QCPortAttributeDefaultValueKey...).
	*/
	
	return nil;
}

+ (QCPlugInExecutionMode)executionMode
{
	/*
	Return the execution mode of the plug-in: kQCPlugInExecutionModeProvider, kQCPlugInExecutionModeProcessor, or kQCPlugInExecutionModeConsumer.
	*/
	
	return kQCPlugInExecutionModeProcessor;
}

+ (QCPlugInTimeMode)timeMode
{
	/*
	Return the time dependency mode of the plug-in: kQCPlugInTimeModeNone, kQCPlugInTimeModeIdle or kQCPlugInTimeModeTimeBase.
	*/
	
	return kQCPlugInTimeModeNone;
}

- (id)init
{
	self = [super init];
	if (self) {
        
        path = [[NSMutableArray alloc]initWithCapacity:10];
	
	}
	
	return self;
}

- (void)finalize
{
	/*
	Release any non garbage collected resources created in -init.
	*/
	
	[super finalize];
}

- (void)dealloc
{
	[path release];
	
	[super dealloc];
}


@end

@implementation viterbiPlugIn (Execution)

- (BOOL)startExecution:(id <QCPlugInContext>)context
{
	/*
	Called by Quartz Composer when rendering of the composition starts: perform any required setup for the plug-in.
	Return NO in case of fatal failure (this will prevent rendering of the composition to start).
	*/
	
	return YES;
}

- (void)enableExecution:(id <QCPlugInContext>)context
{
	/*
	Called by Quartz Composer when the plug-in instance starts being used by Quartz Composer.
	*/
}

- (BOOL)execute:(id <QCPlugInContext>)context atTime:(NSTimeInterval)time withArguments:(NSDictionary *)arguments
{
    NSString *  state1 = [NSString stringWithFormat:@"%@",self.input_state_1];
    const char*s_1 =  [state1 cStringUsingEncoding:1]; 
    
    NSString *  state2 = [NSString stringWithFormat:@"%@",self.input_state_2];
    const char*s_2 =  [state2 cStringUsingEncoding:1]; 
    
    NSString *  state3 = [NSString stringWithFormat:@"%@",self.input_state_3];
    const char*s_3 =  [state3 cStringUsingEncoding:1]; 
    
    NSString *  state4 = [NSString stringWithFormat:@"%@",self.input_state_4];
    const char*s_4 =  [state4 cStringUsingEncoding:1]; 
 
    NSString *  obs1 = [NSString stringWithFormat:@"%@",self.input_observation_1];
    const char*o_1 =  [obs1 cStringUsingEncoding:1]; 
    
    NSString *  obs2 = [NSString stringWithFormat:@"%@",self.input_observation_2];
    const char*o_2 =  [obs2 cStringUsingEncoding:1]; 
    
    NSString *  obs3 = [NSString stringWithFormat:@"%@",self.input_observation_3];
    const char*o_3 =  [obs3 cStringUsingEncoding:1]; 
  
    NSString *  obs4 = [NSString stringWithFormat:@"%@",self.input_observation_4];
    const char*o_4 =  [obs4 cStringUsingEncoding:1];
    

    init_variables(s_1, s_2, s_3, s_4, o_1, o_2, o_3, o_4, self.input_state_1_prob, self.input_state_2_prob,  self.input_state_3_prob, self.input_state_4_prob);
    
    
    if(self.input_run)
        {
        [path removeAllObjects];
        
        
        for (int i = 0; i <5; i++)
            {
                //
            NSString * p = [NSString stringWithCString:(const char * )forward_viterbi(observations, states,  start_probability,  transition_probability, emission_probability)[i].c_str() encoding:1];
            
                   [path  addObject:p];
        
            }
        
               NSLog(@"predict path: %@ ", path);
                self.output_v_path = path;
        }
    
      
	
	return YES;
}

- (void)disableExecution:(id <QCPlugInContext>)context
{
	/*
	Called by Quartz Composer when the plug-in instance stops being used by Quartz Composer.
	*/
}

- (void)stopExecution:(id <QCPlugInContext>)context
{
	/*
	Called by Quartz Composer when rendering of the composition stops: perform any required cleanup for the plug-in.
	*/
}

@end
